# @Author: chunyang.xu
# @Email:  398745129@qq.com
# @Date:   2020-06-08 11:55:54
# @Last Modified time: 2021-08-29 13:30:07
# @github: https://github.com/longfengpili

# !/usr/bin/env python3
# -*- coding:utf-8 -*-

import os
import time

from .base import DBbase
from pydbapi.sql import SqlFileParse

import logging
dblogger = logging.getLogger(__name__)


class DBFileExec(DBbase):

    def __init__(self):
        super(DBFileExec, self).__init__()

    def get_filesqls(self, filepath, **kw):
        sqlfileparser = SqlFileParse(filepath)
        arguments, sqls = sqlfileparser.get_filesqls(**kw)
        return arguments, sqls

    def file_exec(self, filepath, ehandling=None, verbose=0, **kw):
        st = time.time()
        results = {}
        filename = os.path.basename(filepath)
        dblogger.info(f"Start Job 【{filename}】".center(80, '='))
        arguments, sqls = self.get_filesqls(filepath, **kw)
        for desc, sql in sqls.items():
            dblogger.info(f">>> START {desc}")
            sqlverbose = verbose or (2 if 'verbose2' in desc else 1
                                     if 'verbose1' in desc or filename.startswith('test')
                                     else 0)
            sqlehandling = ehandling or ('pass' if 'epass' in desc else 'raise')
            rows, action, result = self.execute(sql, ehandling=sqlehandling, verbose=sqlverbose)
            results[desc] = result
            # dblogger.info(f"End {desc}")
        et = time.time()
        dblogger.info(f"End Job 【{filename}】, cost {et - st:.2f} seconds".center(80, '='))
        return results



# @Author: chunyang.xu
# @Email:  398745129@qq.com
# @Date:   2020-06-02 18:46:58
# @Last Modified time: 2022-01-20 10:00:10
# @github: https://github.com/longfengpili

# !/usr/bin/env python3
# -*- coding:utf-8 -*-

import re
import pandas as pd
from tqdm import tqdm

from pydbapi.sql import SqlParse, SqlCompile
from pydbapi.conf import AUTO_RULES

import logging
dblogger = logging.getLogger(__name__)


class DBbase(object):

    def __init__(self, *args, **kwargs):
        pass

    def get_conn(self):
        pass

    def __execute_step(self, cursor, sql, ehandling='raise'):
        '''[summary]

        [description]
            在conn中执行单步sql
        Arguments:
            cursor {[cursor]} -- [游标]
            sql {[str]} -- [sql]

        Raises:
            ValueError -- [sql执行错误原因及SQL]
        '''
        sql = re.sub(r'\s{2,}', '\n', sql)
        try:
            cursor.execute(sql)
        except Exception as e:
            dblogger.error(f"【Error】{e}, 【error Sql】: {sql}")

            if ehandling == 'raise':
                raise ValueError(f"【Error】:{e}【Sql】:{sql}")

    def cur_results(self, cursor, count):
        results = cursor.fetchmany(count) if count else cursor.fetchall()
        results = list(results) if results else []
        return results

    def cur_columns(self, cursor):
        desc = cursor.description
        columns = tuple(map(lambda x: x[0].lower(), desc)) if desc else None  # 列名

        return desc, columns

    def execute(self, sql, count=None, ehandling='raise', verbose=0):
        '''[summary]

        [description]
            执行sql
        Arguments:
            sql {[str]} -- [sql]

        Keyword Arguments:
            count {[int]} -- [返回的结果数量] (default: {None})
            ehandling {[str]} -- [错误处理] （raise: 错误弹出异常）
            verbose {[int]} -- [打印进程状态] （0：不打印， 1：文字进度， 2：进度条）

        Returns:
            rows {[int]} -- [影响的行数]
            results {[list]} -- [返回的结果]
        '''
        # def cur_getresults(cur, count):
        #     results = cur.fetchmany(count) if count else cur.fetchall()
        #     results = list(results) if results else []
        #     columns = tuple(map(lambda x: x[0].lower(), cur.description)) if cur.description  # 列名
        #     return columns, results

        rows = 0
        idx = 0
        conn = self.get_conn()
        # dblogger.info(conn)
        cur = conn.cursor()
        sql = sql if sql.strip().endswith(';') else sql.strip() + ';'
        sqls = sql.split(";")[:-1]
        sqls = [sql.strip() for sql in sqls if sql]
        sqls_length = len(sqls)
        bar_format = '{l_bar}{bar:30}| {n_fmt}/{total_fmt} [{elapsed}<{remaining}] {postfix[0]}'
        sqls = sqls if verbose <= 1 else tqdm(sqls, ncols=100, postfix=['START'], bar_format=bar_format)  # 如果verbose>=2则显示进度条
        for _sql in sqls:
            results = None
            idx += 1
            # dblogger.info(_sql)
            parser = SqlParse(_sql)
            comment, sql, action, tablename = parser.comment, parser.sql, parser.action, parser.tablename
            if not sql:
                # dblogger.info(f'【{idx:0>2d}_PROGRESS】 no run !!!\n{_sql}')
                continue

            step = f"【{idx:0>2d}_PROGRESS】({action}){tablename}::{comment}"
            if verbose == 1:
                dblogger.info(f"{step}")
            elif verbose >= 2:
                sqls.postfix[0] = f"{step}"
                sqls.update()
            else:
                pass

            self.__execute_step(cur, sql, ehandling=ehandling)

            if action == 'SELECT' and (verbose or idx == sqls_length):
                # columns, results = cur_getresults(cur, count)
                results = self.cur_results(cur, count)
                desc, columns = self.cur_columns(cur)
                if verbose and columns:
                    dblogger.info(f"\n{pd.DataFrame(results, columns=columns)}")
                elif not columns:
                    dblogger.warning(f"Not Columns, cursor description is {desc}")
                else:
                    pass

                if columns:
                    results.insert(0, columns)
        try:
            conn.commit()
        except Exception as e:
            dblogger.error(e)
            conn.rollback()
        rows = cur.rowcount
        conn.close()
        return rows, action, results


class DBCommon(DBbase):

    def __init__(self):
        self.auto_rules = AUTO_RULES
        super(DBCommon, self).__init__()

    def __check_isauto(self, tablename):
        '''[summary]

        [description]
            通过tablename控制是否可以通过python代码处理
        Arguments:
            tablename {[str]} -- [表名]
        '''
        if not self.auto_rules:
            return True
        for rule in self.auto_rules:
            if rule in tablename:
                return True
        else:
            raise Exception(f"【drop】 please drop [{tablename}] on workbench! Or add rule into auto_rules !")
        return False

    def drop(self, tablename, verbose=0):
        if self.__check_isauto(tablename):
            sqlcompile = SqlCompile(tablename)
            sql_for_drop = sqlcompile.drop()
            rows, action, result = self.execute(sql_for_drop, verbose=verbose)
            dblogger.info(f'【{action}】{tablename} drop succeed !')
            return rows, action, result

    def delete(self, tablename, condition, verbose=0):
        if self.__check_isauto(tablename):
            sqlcompile = SqlCompile(tablename)
            sql_for_delete = sqlcompile.delete(condition)
            rows, action, result = self.execute(sql_for_delete, verbose=verbose)
            dblogger.info(f'【{action}】{tablename} delete {rows} rows succeed !')
            return rows, action, result

    def insert(self, tablename, columns, inserttype='value', values=None, chunksize=1000, 
               fromtable=None, condition=None, verbose=0):
        if values:
            vlength = len(values)

        if self.__check_isauto(tablename):
            sqlcompile = SqlCompile(tablename)
            sql_for_insert = sqlcompile.insert(columns, inserttype=inserttype, values=values,
                                               chunksize=chunksize, fromtable=fromtable, condition=condition)
            rows, action, result = self.execute(sql_for_insert, verbose=verbose)

            if values and rows != vlength % chunksize:
                raise Exception(f'Insert Error !!!')

            rows = vlength if values else rows
            dblogger.info(f'【{action}】{tablename} insert {rows} rows succeed !')
            return rows, action, result

    def get_columns(self, tablename, verbose=0):
        sql = f"select * from {tablename} limit 1;"
        rows, action, result = self.execute(sql, verbose=verbose)
        columns = result[0]
        columns = [c.lower() for c in columns]
        return columns

    def select(self, tablename, columns, condition=None, verbose=0):
        '''[summary]

        [description]
            执行select 
        Arguments:
            tablename {[str]} -- [表名]
            columns {[dict]} -- [列的信息]

        Keyword Arguments:
            condition {[str]} -- [where中的表达式] (default: {None})

        Returns:
            rows[int] -- [影响的数量]
            action[str] -- [sql表达式DML]
            result[list] -- [结果, 第一个元素是列名]
        '''
        sqlcompile = SqlCompile(tablename)
        sql_for_select = sqlcompile.select_base(columns, condition=condition)
        rows, action, result = self.execute(sql_for_select, verbose=verbose)
        return rows, action, result

    def add_columns(self, tablename, columns, verbose=0):
        old_columns = self.get_columns(tablename)
        old_columns = set(old_columns)
        new_columns = columns.new_cols
        new_columns = set([col.strip() for col in new_columns.split(',')])
        # dblogger.info(f'{old_columns}, {new_columns}')

        if old_columns == new_columns:
            dblogger.info(f'【{tablename}】columns not changed !')
        if old_columns - new_columns:
            raise Exception(f"【{tablename}】columns【{old_columns - new_columns}】 not set, should exists !")
        if new_columns - old_columns:
            sqlcompile = SqlCompile(tablename)
            add_columns = new_columns - old_columns
            for col_name in add_columns:
                column = columns.get_column_by_name(col_name)
                sql = sqlcompile.add_column(column.newname, column.coltype)
                self.execute(sql, verbose=0)
            dblogger.info(f'【{tablename}】add columns succeeded !【{new_columns - old_columns}】')



# @Author: chunyang.xu
# @Email:  398745129@qq.com
# @Date:   2020-06-03 14:04:33
# @Last Modified time: 2021-08-25 16:46:12
# @github: https://github.com/longfengpili

# !/usr/bin/env python3
# -*- coding:utf-8 -*-


from pydbapi.col import ColumnsModel


class SqlCompile(object):

    def __init__(self, tablename):
        self.tablename = tablename

    def select_base(self, columns, fromtable=None, condition=None):
        '''[summary]

        [description]
            生成select sql (未考虑join，所以暂时用base)
        Arguments:
            columns {[ColumnsModel]} -- [列信息需要按照排列顺序处理]

        Keyword Arguments:
            condition {[条件]} -- [where中的条件] (default: {None})

        Returns:
            [str] -- [返回sql]

        Raises:
            TypeError -- [检查columns的情况]
        '''
        tablename = fromtable or self.tablename
        if not isinstance(columns, ColumnsModel):
            raise TypeError("colums must be a ColumnsModel !")

        sql = f'select {columns.select_cols}\nfrom {tablename}'
        condition = f"where {condition}" if condition else ''
        group = f'group by {columns.group_cols}' if columns.group_cols else ''
        order = f'order by {columns.order_cols}' if columns.order_cols else ''

        if condition:
            sql = sql + '\n' + condition
        if group:
            sql = sql + '\n' + group
        if order:
            sql = sql + '\n' + order
        sql = sql + '\n;'
        return sql

    def create_nonindex(self, columns):
        '''[summary]

        [description]
            create sql
        Arguments:
            self.tablename {[str]} -- [表名]
            columns {[ColumnsModel]} -- [列信息]

        Returns:
            [str] -- [sql]

        Raises:
            TypeError -- [类别错误]
        '''

        if not isinstance(columns, ColumnsModel):
            raise TypeError("colums must be a ColumnsModel !")

        sql = f'create table if not exists {self.tablename}\n{columns.create_cols};'
        return sql

    def drop(self):
        sql = f'drop table if exists {self.tablename};'
        return sql

    def _insert_value(self, columns, values):
        '''[summary]

        [description]
            插入数据
        Arguments:
            self.tablename {[str]} -- [表名]
            columns {[ColumnsModel]} -- [列信息]
            values {[list]} -- [插入的数据]
        '''
        def deal_values(values):
            j_values = []
            if not values:
                raise ValueError(f"{values} is empty !!!")
            if not isinstance(values, list):
                raise TypeError('values must be a list !')
            j_values = [str(tuple(value)) for value in values]
            j_values = ',\n'.join(j_values)
            # j_values = j_values.replace('\'\'', 'null').replace('None', 'null')  # 空值替换为null
            return j_values

        values = deal_values(values)

        if not isinstance(columns, ColumnsModel):
            raise TypeError("colums must be a ColumnsModel !")

        sql = f'insert into {self.tablename}\n({columns.new_cols})\nvalues\n{values};'
        sql = sql.replace('\'\'', 'Null').replace('None', 'Null')  # 空值替换为null
        return sql

    def _insert_by_value(self, columns, values, chunksize=1000):
        sql = []
        vlength = len(values)
        step = 0
        for i in range(0, vlength, chunksize):
            step += 1
            maxi = i+chunksize
            maxi = maxi if maxi < vlength else vlength
            chvalues = values[i: maxi]
            _sql = self._insert_value(columns, chvalues)
            _sql = f"--[NO.{step}]insert {maxi-i} rows, From [{i+1}] to [{maxi}]\n{_sql}"
            sql.append(_sql)

        sql = ''.join(sql)
        # print(sql)
        return sql

    def _insert_by_select(self, fromtable, columns, condition=None):
        selectsql = self.select_base(columns, fromtable, condition=condition)

        sql = f'''insert into {self.tablename}\n({columns.new_cols})\n{selectsql}'''
        return sql

    def insert(self, columns, inserttype='value', values=None, chunksize=1000, fromtable=None, condition=None):
        if inserttype == 'value':
            if not values:
                raise Exception(f"InsertType is {inserttype}, values must be not None")
            sql = self._insert_by_value(columns, values, chunksize=chunksize)
            # print(sql)
        elif inserttype == 'select':
            if not fromtable:
                raise Exception(f"InsertType is {inserttype}, fromtable must be not None")
            sql = self._insert_by_select(fromtable, columns, condition=condition)
        else:
            raise Exception(f"Not supported {inserttype}, insertType must be value or select")
        return sql

    def delete(self, condition):
        sql = f'''delete from {self.tablename} where {condition};'''
        return sql

    def add_column(self, colname, coltype):
        coltype = f"{coltype}(32)" if coltype == 'varchar' else coltype
        sql = f'alter table {self.tablename} add column {colname} {coltype};'
        return sql



# @Author: chunyang.xu
# @Email:  398745129@qq.com
# @Date:   2020-06-04 10:58:54
# @Last Modified time: 2020-12-01 18:42:13
# @github: https://github.com/longfengpili

# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# from pydbapi.api import SqliteDB, RedshiftDB, MysqlDB, SnowflakeDB
# from pydbapi.sql import SqlParse, SqlCompile, SqlFileParse, ColumnModel, ColumnsModel

# __all__ = ['SqliteDB', 'RedshiftDB', 'MysqlDB', 'SnowflakeDB',
#            'SqlParse', 'SqlCompile', 'SqlFileParse', 'ColumnModel', 'ColumnsModel']



# @Author: chunyang.xu
# @Email:  398745129@qq.com
# @Date:   2020-06-03 10:31:52
# @Last Modified time: 2020-06-17 21:08:30
# @github: https://github.com/longfengpili

#!/usr/bin/env python3
# -*- coding:utf-8 -*-

from .base import DBCommon
from .fileexec import DBFileExec

__all__ = ['DBCommon', 'DBFileExec']



# # @Author: chunyang.xu
# # @Email:  398745129@qq.com
# # @Date:   2020-10-22 16:12:47
# # @Last Modified time: 2021-03-08 15:09:12
# # @github: https://github.com/longfengpili

# # !/usr/bin/env python3
# # -*- coding:utf-8 -*-

# import threading
# import snowflake.connector

# from pydbapi.db import DBCommon, DBFileExec
# from pydbapi.sql import SqlCompile
# from pydbapi.conf import AUTO_RULES


# import logging
# import logging.config
# from pydbapi.conf import LOGGING_CONFIG
# logging.config.dictConfig(LOGGING_CONFIG)
# sflogger = logging.getLogger('snowflake')


# class SqlSnowflakeCompile(SqlCompile):
#     '''[summary]

#     [description]
#         构造mysql sql
#     Extends:
#         SqlCompile
#     '''

#     def __init__(self, tablename):
#         super(SqlSnowflakeCompile, self).__init__(tablename)

#     def create(self, columns, indexes):
#         sql = self.create_nonindex(columns)
#         if indexes and not isinstance(indexes, list):
#             raise TypeError(f"indexes must be a list !")
#         if indexes:
#             indexes = ','.join(indexes)
#             sql = f"{sql.replace(';', '')}\ncluster by ({indexes});"
#         return sql


# class SnowflakeDB(DBCommon, DBFileExec):
#     _instance_lock = threading.Lock()

#     def __init__(self, user, password, account, warehouse, database, schema, safe_rule=True):
#         self.user = user
#         self.password = password
#         self.account = account
#         self.warehouse = warehouse
#         self.database = database
#         self.schema = schema
#         super(SnowflakeDB, self).__init__()
#         self.auto_rules = AUTO_RULES if safe_rule else None

#     @classmethod
#     def get_instance(cls, *args, **kwargs):
#         if not hasattr(SnowflakeDB, '_instance'):
#             with SnowflakeDB._instance_lock:
#                 if not hasattr(SnowflakeDB, '_instance'):
#                     SnowflakeDB._instance = SnowflakeDB(*args, **kwargs)

#         return SnowflakeDB._instance

#     def get_conn(self):
#         conn = snowflake.connector.connect(database=self.database, user=self.user, password=self.password,
#                                            account=self.account, warehouse=self.warehouse, schema=self.schema)
#         if not conn:
#             self.get_conn()
#         return conn

#     def create(self, tablename, columns, indexes=None, verbose=0):
#         # tablename = f"{self.database}.{tablename}"
#         sqlcompile = SqlSnowflakeCompile(tablename)
#         sql_for_create = sqlcompile.create(columns, indexes)
#         rows, action, result = self.execute(sql_for_create, verbose=verbose)
#         return rows, action, result



# @Author: chunyang.xu
# @Email:  398745129@qq.com
# @Date:   2020-06-03 10:31:36
# @Last Modified time: 2021-02-01 19:18:38
# @github: https://github.com/longfengpili

# !/usr/bin/env python3
# -*- coding:utf-8 -*-

from .parse import SqlParse, SqlFileParse
from .compile import SqlCompile

__all__ = ['SqlParse', 'SqlCompile', 'SqlFileParse']



# @Author: chunyang.xu
# @Email:  398745129@qq.com
# @Date:   2020-06-10 14:14:53
# @Last Modified time: 2022-01-20 09:47:50
# @github: https://github.com/longfengpili

# !/usr/bin/env python3
# -*- coding:utf-8 -*-

import re
import os
import sys
import colorlog

AUTO_RULES = ['test_xu']  # 可以自动执行表名（表名包含即可）
REDSHIFT_AUTO_RULES = AUTO_RULES + ['_data_aniland']  # Amazon Redshift 可以自动执行表名（表名包含即可）

# logging settings
USERPATH = os.environ['USERPROFILE'] if 'USERPROFILE' in os.environ else os.environ['HOME'] if 'HOME' in os.environ else ''
LOG_BASE_PATH = os.path.join(USERPATH, '.pydbapilog')  # 可以user目录下查看日志
PROJECT_NAME = re.sub(':?\\\\', '_', os.getcwd())
LOGGING_CONFIG = {
    'version': 1,  # 保留字
    'disable_existing_loggers': False,  # 禁用已经存在的logger实例
    # 日志文件的格式
    'formatters': {
        # 详细的日志格式
        'standard': {
            'format': '%(asctime)s.%(msecs)03d - %(threadName)s:%(thread)d - %(name)s - %(levelname)s - %(pathname)s - %(lineno)d - %(message)s',
            'datefmt': '%Y-%m-%d %H:%M:%S',
        },
        # 简单的日志格式
        'simple': {
            'format': '%(asctime)s.%(msecs)03d - %(threadName)s - %(name)s - %(levelname)s - %(filename)s - %(lineno)d - %(message)s',
            'datefmt': '%Y-%m-%d %H:%M:%S',
        },
        # 定义一个特殊的日志格式
        'collect': {
            'format': '%(message)s'
        },
        # color
        'color': {
            '()': colorlog.ColoredFormatter,
            'format': '%(asctime)s.%(msecs)03d - %(threadName)s - %(name)s - %(levelname)s - %(filename)s - %(lineno)d - %(log_color)s%(message)s',
            'datefmt': '%Y-%m-%d %H:%M:%S',
            'log_colors': {
                'CRITICAL': 'bold_red',
                'ERROR': 'red',
                'WARNING': 'purple',
                'INFO': 'green',
                'DEBUG': 'yellow'
            }
        }
    },
    # 过滤器
    'filters': {
    },
    # 处理器
    'handlers': {
        # 在终端打印
        'console': {
            'level': 'INFO',
            'filters': [],
            'class': 'logging.StreamHandler',  #
            'formatter': 'color' if sys.stdout.isatty() else 'simple'
        },
        # 默认的
        'default': {
            'level': 'INFO',
            'class': 'pydbapi.conf.MakeFileHandler',  # 能够判断创建日持文件
            'filename': os.path.join(LOG_BASE_PATH, f'{PROJECT_NAME}_default.log'),  # 日志文件
            'when': 'd',  # 每天备份
            'interval': 1,
            'backupCount': 30,  # 最多备份几个
            'formatter': 'standard',
            'encoding': 'utf-8',
        },
        'db': {
            'level': 'INFO',
            'class': 'pydbapi.conf.MakeFileHandler',  # 保存到文件，自动切
            'filename': os.path.join(LOG_BASE_PATH, f'{PROJECT_NAME}_db.log'),  # 日志文件
            'when': 'd',  # 每小时备份
            'interval': 1,
            'backupCount': 30,
            'formatter': 'simple',
            'encoding': "utf-8"
        },
        'sql': {
            'level': 'INFO',
            'class': 'logging.handlers.TimedRotatingFileHandler',  # 保存到文件，自动切
            'filename': os.path.join(LOG_BASE_PATH, f'{PROJECT_NAME}_sql.log'),  # 日志文件
            'when': 'd',  # 每小时备份
            'interval': 1,
            'backupCount': 30,
            'formatter': 'simple',
            'encoding': "utf-8"
        },
    },
    'loggers': {
        # 默认的logger应用如下配置
        '': {
            'handlers': ['console', 'default'],
            'level': 'INFO',
            'propagate': True,  # 向不向更高级别的logger传递
        },
        'db': {
            'handlers': ['console', 'db'],
            'level': 'INFO',
            'propagate': False,  # 向不向更高级别的logger传递
        },
        'sql': {
            'handlers': ['console', 'sql'],
            'level': 'INFO',
            'propagate': False,  # 向不向更高级别的logger传递
        },
        'redshift': {
            'handlers': ['console', 'db'],
            'level': 'INFO',
            'propagate': False,  # 向不向更高级别的logger传递
        },
        'sqlite': {
            'handlers': ['console', 'db'],
            'level': 'INFO',
            'propagate': False,  # 向不向更高级别的logger传递
        },
        'mysql': {
            'handlers': ['console', 'db'],
            'level': 'INFO',
            'propagate': False,  # 向不向更高级别的logger传递
        },
        'snowflake': {
            'handlers': ['console', 'db'],
            'level': 'INFO',
            'propagate': False,  # 向不向更高级别的logger传递
        },
    },
}



# @Author: chunyang.xu
# @Email:  398745129@qq.com
# @Date:   2020-11-30 16:28:21
# @Last Modified time: 2021-12-30 19:45:56
# @github: https://github.com/longfengpili

# !/usr/bin/env python3
# -*- coding:utf-8 -*-


class ColumnModel(object):

    def __init__(self, newname, coltype='varchar', sqlexpr=None, func=None, order=0):
        self.newname = newname
        self.coltype = coltype
        self.sqlexpr = sqlexpr or newname
        self.func = self.__check_func(func)
        self.order = order

    def __repr__(self):
        return f"{self.newname}({self.coltype})"

    def __contains__(self, newname):
        return newname == self.newname

    def __check_func(self, func):
        if func and func not in ['min', 'max', 'sum', 'count']:
            raise ValueError(f"func:{func} not supported!")
        return func

    @property
    def final_sqlexpr(self):
        if self.func:
            final_sqlexpr = f"{self.func}({self.sqlexpr}) as {self.newname}"
        elif self.newname != self.sqlexpr:
            final_sqlexpr = f"{self.sqlexpr} as {self.newname}"
        else:
            final_sqlexpr = self.newname

        return final_sqlexpr

    @property
    def create_sqlexpr(self):
        create_sqlexpr = f"{self.newname} {self.coltype}"
        return create_sqlexpr


class ColumnsModel(object):

    def __init__(self, *columns):
        self.columns = columns

    def __repr__(self):
        return f"{[column for column in self.columns]}"

    def __getitem__(self, key):
        return self.columns[key]

    def index(self, key):
        return self.columns[key]

    def __contains__(self, name):
        col = self.get_column_by_name(name)
        isin = True if col else False
        return isin

    @property
    def func_cols(self):
        func_cols = [col for col in self.columns if col.func]
        return func_cols

    @property
    def nonfunc_cols(self):
        nonfunc_cols = [col for col in self.columns if not col.func]
        return nonfunc_cols

    @property
    def new_cols(self):
        all_cols = self.nonfunc_cols + self.func_cols
        new_cols = [col.newname for col in all_cols]
        new_cols = ', '.join(new_cols)
        return new_cols

    @property
    def create_cols(self):
        create_cols = [col.create_sqlexpr for col in self.columns]
        create_cols = '(' + ',\n'.join(create_cols) + ')'
        return create_cols

    @property
    def select_cols(self):
        all_cols = self.nonfunc_cols + self.func_cols
        all_cols = [col.final_sqlexpr for col in all_cols]
        select_cols = ',\n'.join(all_cols)
        return select_cols

    @property
    def group_cols(self):
        if self.func_cols:
            group_cols = [col.newname for col in self.nonfunc_cols]
            group_cols = ', '.join(group_cols)
            return group_cols

    @property
    def order_cols(self):
        all_cols = self.nonfunc_cols + self.func_cols
        order_cols = [(idx+1, col) for idx, col in enumerate(all_cols) if col.order > 0]
        order_cols_sorted = sorted(order_cols, key=lambda x: [x[1].order, x[1].newname])
        order_cols = [f"{col[0]}" for col in order_cols_sorted]
        order_cols = ', '.join(order_cols)
        return order_cols

    def get_column_by_name(self, name):
        for col in self.columns:
            if col.newname == name:
                return col



# @Author: chunyang.xu
# @Email:  398745129@qq.com
# @Date:   2020-06-10 14:40:50
# @Last Modified time: 2021-11-22 14:30:29
# @github: https://github.com/longfengpili

# !/usr/bin/env python3
# -*- coding:utf-8 -*-


import threading
import psycopg2

from pydbapi.db import DBCommon, DBFileExec
from pydbapi.sql import SqlCompile
from pydbapi.conf import REDSHIFT_AUTO_RULES


import logging
redlogger = logging.getLogger(__name__)


class SqlRedshiftCompile(SqlCompile):
    '''[summary]

    [description]
        构造redshift sql
    Extends:
        SqlCompile
    '''

    def __init__(self, tablename):
        super(SqlRedshiftCompile, self).__init__(tablename)

    def create(self, columns, indexes):
        sql = self.create_nonindex(columns)
        if indexes and not isinstance(indexes, list):
            raise TypeError(f"indexes must be a list !")
        if indexes:
            indexes = ','.join(indexes)
            sql = f"{sql.replace(';', '')}\ninterleaved sortkey({indexes});"
        return sql


class RedshiftDB(DBCommon, DBFileExec):
    _instance_lock = threading.Lock()

    def __init__(self, host, user, password, database, port='5439', safe_rule=True):
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.database = database
        super(RedshiftDB, self).__init__()
        self.auto_rules = REDSHIFT_AUTO_RULES if safe_rule else None

    def __new__(cls, *args, **kwargs):
        if not hasattr(RedshiftDB, '_instance'):
            with RedshiftDB._instance_lock:
                if not hasattr(RedshiftDB, '_instance'):
                    RedshiftDB._instance = super().__new__(cls)

        return RedshiftDB._instance

    # @classmethod
    # def get_instance(cls, *args, **kwargs):
    #     if not hasattr(RedshiftDB, '_instance'):
    #         with RedshiftDB._instance_lock:
    #             if not hasattr(RedshiftDB, '_instance'):
    #                 RedshiftDB._instance = cls(*args, **kwargs)

    #     return RedshiftDB._instance

    def get_conn(self):
        conn = psycopg2.connect(database=self.database, user=self.user, password=self.password, host=self.host, port=self.port)
        if not conn:
            self.get_conn()
        return conn

    def create(self, tablename, columns, indexes=None, verbose=0):
        # tablename = f"{self.database}.{tablename}"
        sqlcompile = SqlRedshiftCompile(tablename)
        sql_for_create = sqlcompile.create(columns, indexes)
        rows, action, result = self.execute(sql_for_create, verbose=verbose)
        return rows, action, result



# -*- coding: utf-8 -*-
# @Author: chunyang.xu
# @Date:   2021-02-01 19:18:20
# @Last Modified by:   chunyang.xu
# @Last Modified time: 2021-02-01 19:18:29


from .colmodel import ColumnModel, ColumnsModel

__all__ = ['ColumnModel', 'ColumnsModel']



# @Author: chunyang.xu
# @Email:  398745129@qq.com
# @Date:   2020-06-10 12:18:59
# @Last Modified time: 2020-06-17 21:08:41
# @github: https://github.com/longfengpili

#!/usr/bin/env python3
# -*- coding:utf-8 -*-

from .myhandlers import MakeFileHandler
from .settings import *

# __all__ = ['MakeFileHandler', 'LOGGING_CONFIG', 'AUTO_RULES', 'REDSHIFT_AUTO_RULES']



# @Author: chunyang.xu
# @Email:  398745129@qq.com
# @Date:   2020-06-10 14:40:50
# @Last Modified time: 2021-12-30 19:36:17
# @github: https://github.com/longfengpili

# !/usr/bin/env python3
# -*- coding:utf-8 -*-

import re
import threading
import pymysql

from pydbapi.db import DBCommon, DBFileExec
from pydbapi.sql import SqlCompile
from pydbapi.conf import AUTO_RULES


import logging
mysqllogger = logging.getLogger(__name__)


class SqlMysqlCompile(SqlCompile):
    '''[summary]

    [description]
        构造mysql sql
    Extends:
        SqlCompile
    '''

    def __init__(self, tablename):
        super(SqlMysqlCompile, self).__init__(tablename)

    def create_indexes(self, columns, indexes, index_part=128, ismultiple=True):
        if indexes and not isinstance(indexes, list):
            raise TypeError(f"indexes must be a list, but got {indexes} !")
        _indexes = []
        for index in indexes:
            column = columns.get_column_by_name(index)
            coltype = column.coltype
            varlength = re.search('varchar\((\d+)\)', coltype)
            varlength = int(varlength.group(1)) if varlength else 128
            index_part = varlength if varlength < index_part else index_part
            index = f"{index}({index_part})" if coltype.startswith('varchar') else index
            _indexes.append(index)

        if ismultiple:
            indexes = ', '.join(_indexes)
            indexes = f"index multiple_index({indexes})"
        else:
            indexes = [f"index {_index}_index({_index})" for _index in indexes]
            indexes = ',\n'.join(indexes)

        return indexes

    def create_partition(self, partition):
        coltype = partition.coltype
        if coltype != 'date':
            raise TypeError(f"{partition} only support date type !")
        partition = f"partition by hash (to_days({partition.newname}))"
        return partition

    def create(self, columns, indexes, index_part=128, ismultiple_index=True, 
               partition=None, distribution=None, isdoris=False):
        'mysql 暂不考虑索引'
        sql = self.create_nonindex(columns)

        if indexes and not isinstance(indexes, list):
            raise TypeError(f"indexes must be a list, but got {indexes} !")

        if indexes:
            indexes = self.create_indexes(columns, indexes, index_part=index_part, ismultiple=ismultiple_index)
            sql = sql.replace(');', f",\n{indexes});")

        if partition:
            partition = columns.get_column_by_name(partition)
            partition = self.create_partition(partition)
            sql = sql.replace(';', f'\n{partition};')

        if isdoris:
            distributed_col = distribution or columns.index(0).newname
            distributed = f"distributed by hash({distributed_col})"
            sql = sql.replace(';', f'\n{distributed};')

        return sql

    def dumpsql(self, columns, dumpfile, fromtable=None, condition=None):
        selectsql = self.select_base(columns, fromtable=fromtable, condition=condition)
        intosql = f'into outfile "{dumpfile}" fields terminated by ",";'
        dumpsql = selectsql.replace(";", intosql)
        return dumpsql

    def loadsql(self, columns, loadfile, intotable=None, fieldterminated=','):
        intotable = intotable or self.tablename
        loadsql = f'''load data infile "{loadfile}" into table {intotable}
                      fields terminated by "{fieldterminated}" ({columns.select_cols});'''
        return loadsql


class MysqlDB(DBCommon, DBFileExec):
    _instance_lock = threading.Lock()

    def __init__(self, host, user, password, database, port=3306, charset="utf8", safe_rule=True, isdoris=False):
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.database = database
        self.charset = charset
        self.isdoris = isdoris
        super(MysqlDB, self).__init__()
        self.auto_rules = AUTO_RULES if safe_rule else None

    def __new__(cls, *args, **kwargs):
        if not hasattr(MysqlDB, '_instance'):
            with MysqlDB._instance_lock:
                if not hasattr(MysqlDB, '_instance'):
                    MysqlDB._instance = super().__new__(cls)

        return MysqlDB._instance

    # @classmethod
    # def get_instance(cls, *args, **kwargs):
    #     mysqllogger.info(MysqlDB._instance_lock)
    #     if not hasattr(MysqlDB, '_instance'):
    #         mysqllogger.info(MysqlDB._instance_lock)
    #         with MysqlDB._instance_lock:
    #             if not hasattr(MysqlDB, '_instance'):
    #                 MysqlDB._instance = cls(*args, **kwargs)

    #     return MysqlDB._instance

    def get_conn(self):
        conn = pymysql.connect(database=self.database, user=self.user, password=self.password,
                               host=self.host, port=self.port, charset=self.charset)
        if not conn:
            self.get_conn()
        return conn

    def create(self, tablename, columns, indexes=None, index_part=128, ismultiple_index=True,
               partition=None, distribution=None, verbose=0):
        # tablename = f"{self.database}.{tablename}"
        sqlcompile = SqlMysqlCompile(tablename)
        sql_for_create = sqlcompile.create(columns, indexes, index_part=index_part,
                                           ismultiple_index=ismultiple_index, partition=partition,
                                           distribution=distribution, isdoris=self.isdoris)
        rows, action, result = self.execute(sql_for_create, verbose=verbose)
        return rows, action, result

    def dumpdata(self, tablename, columns, dumpfile, condition=None, verbose=0):
        sqlcompile = SqlMysqlCompile(tablename)
        sql_for_dump = sqlcompile.dumpsql(columns, dumpfile, condition=condition)
        rows, action, result = self.execute(sql_for_dump, verbose=verbose)
        mysqllogger.info(f"【{action}】{tablename} dumpdata {rows} rows succeed, outfile: {dumpfile} !")
        return rows, action, result

    def loaddata(self, tablename, columns, loadfile, fieldterminated=',', verbose=0):
        sqlcompile = SqlMysqlCompile(tablename)
        sql_for_load = sqlcompile.loadsql(columns, loadfile, fieldterminated=fieldterminated)
        rows, action, result = self.execute(sql_for_load, verbose=verbose)
        mysqllogger.info(f"【{action}】{tablename} loaddata {rows} rows succeed, loadfile: {loadfile} !")
        return rows, action, result



# @Author: chunyang.xu
# @Email:  398745129@qq.com
# @Date:   2020-06-09 16:46:54
# @Last Modified time: 2022-01-20 10:15:32
# @github: https://github.com/longfengpili

# !/usr/bin/env python3
# -*- coding:utf-8 -*-

import os
import sys
import shutil
import setuptools

VERSION = '0.0.84'
PROJECT_NAME = 'pydbapi'

with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

with open('./requirements.txt', 'r', encoding='utf-8') as f:
    requires = f.readlines()

if sys.argv[1] == 'build':
    os.system('python setup.py sdist bdist_wheel')
elif sys.argv[1] == 'publish':
    os.system('python setup.py sdist bdist_wheel')
    os.system(f'python -m twine upload dist/*{VERSION}* --repository {PROJECT_NAME}')
    # shutil.rmtree(f'./{PROJECT_NAME}-{VERSION}')
    shutil.rmtree('./build')
    shutil.rmtree(f'./{PROJECT_NAME}.egg-info')
    sys.exit()

setuptools.setup(
    name=PROJECT_NAME,  # Replace with your own username
    version=VERSION,
    author="longfengpili",
    author_email="398745129@qq.com",
    description="A simple database API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=f"https://pypi.org/project/{PROJECT_NAME}/",
    packages=setuptools.find_packages(exclude=["tests.*", "tests"]),
    install_requires=requires,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
    keywords=["dbapi", "sqlite3", "redshift", 'snowflake'],
    python_requires=">=3.6",
    project_urls={
        'Documentation': f'https://github.com/longfengpili/{PROJECT_NAME}/blob/master/README.md',
        'Source': f'https://github.com/longfengpili/{PROJECT_NAME}',
    },
)



# @Author: chunyang.xu
# @Email:  398745129@qq.com
# @Date:   2020-06-08 14:02:33
# @Last Modified time: 2021-03-08 15:09:19
# @github: https://github.com/longfengpili

#!/usr/bin/env python3
# -*- coding:utf-8 -*-

from .redshift import RedshiftDB
from .sqlite import SqliteDB
from .mysql import MysqlDB
# from .snowflake import SnowflakeDB

__doc__ = "数据库接口"
__all__ = ['RedshiftDB', 'SqliteDB', 'MysqlDB']



# @Author: chunyang.xu
# @Email:  398745129@qq.com
# @Date:   2020-06-08 16:42:55
# @Last Modified time: 2021-03-26 14:43:20
# @github: https://github.com/longfengpili

# !/usr/bin/env python3
# -*- coding:utf-8 -*-

import logging.handlers
import os
import errno


def mklogdir(path):
    try:
        os.makedirs(path, exist_ok=True)  # Python>3.2
    except TypeError:
        try:
            os.makedirs(path)
        except OSError as exc:  # Python >2.5
            if exc.errno == errno.EEXIST and os.path.isdir(path):
                pass
            else: raise


class MakeFileHandler(logging.handlers.TimedRotatingFileHandler):
    def __init__(self, filename, when='h', interval=1, backupCount=0, encoding=None, delay=False, utc=False, atTime=None):
        mklogdir(os.path.dirname(filename))
        super(MakeFileHandler, self).__init__(filename, when, interval, backupCount, encoding, delay, utc, atTime)



# @Author: chunyang.xu
# @Email:  398745129@qq.com
# @Date:   2020-06-03 15:25:44
# @Last Modified time: 2021-11-22 14:30:21
# @github: https://github.com/longfengpili

# !/usr/bin/env python3
# -*- coding:utf-8 -*-

import os
import threading
import sqlite3

from pydbapi.db import DBCommon, DBFileExec
from pydbapi.sql import SqlCompile

import logging
sqlitelogger = logging.getLogger(__name__)


USERPATH = os.environ['USERPROFILE'] if 'USERPROFILE' in os.environ else os.environ['HOME'] if 'HOME' in os.environ else ''


class SqliteCompile(SqlCompile):
    '''[summary]

    [description]
        构造redshift sql
    Extends:
        SqlCompile
    '''
    def __init__(self, tablename):
        super(SqliteCompile, self).__init__(tablename)

    def create(self, columns, indexes):
        'sqlite 暂不考虑索引'
        sql = self.create_nonindex(columns)
        # if indexes and not isinstance(indexes, list):
        #     raise TypeError(f"indexes must be a list !")
        # if indexes:
        #     indexes = ','.join(indexes)
        #     sql = f"{sql.replace(';', '')}interleaved sortkey({indexes});"
        return sql


class SqliteDB(DBCommon, DBFileExec):
    _instance_lock = threading.Lock()

    def __init__(self, database=None):
        self.database = database if database else os.path.join(USERPATH, 'sqlite3_test.db')
        super(SqliteDB, self).__init__()

    def __new__(cls, *args, **kwargs):
        if not hasattr(SqliteDB, '_instance'):
            with SqliteDB._instance_lock:
                if not hasattr(SqliteDB, '_instance'):
                    SqliteDB._instance = super().__new__(cls)

        return SqliteDB._instance

    # @classmethod
    # def get_instance(cls, *args, **kwargs):
    #     if not hasattr(SqliteDB, '_instance'):
    #         with SqliteDB._instance_lock:
    #             if not hasattr(SqliteDB, '_instance'):
    #                 SqliteDB._instance = cls(*args, **kwargs)

    #     return SqliteDB._instance

    def get_conn(self):
        conn = sqlite3.connect(database=self.database)
        if not conn:
            self.get_conn()
        return conn

    def create(self, tablename, columns, indexes=None, verbose=0):
        # tablename = f"{self.database}.{tablename}"
        sqlcompile = SqliteCompile(tablename)
        sql_for_create = sqlcompile.create(columns, indexes)
        rows, action, result = self.execute(sql_for_create, verbose=verbose)
        return rows, action, result



# @Author: chunyang.xu
# @Email:  398745129@qq.com
# @Date:   2020-06-03 10:51:08
# @Last Modified time: 2021-09-23 11:37:11
# @github: https://github.com/longfengpili

#!/usr/bin/env python3
# -*- coding:utf-8 -*-

import re
import os

from datetime import datetime, date, timedelta

import logging
sqllogger = logging.getLogger(__name__)


REG_BEHIND = r'(?=[,();:\s])'


class SqlParse(object):

    def __init__(self, orisql):
        self.orisql = orisql

    @property
    def purpose(self):
        purpose = re.match('-- *(【.*?】)\n', self.orisql.strip())
        purpose = f"{purpose.group(1)}" if purpose else 'No Description'
        return purpose

    @property
    def sql(self):
        sql = re.sub('^(--.*?\n){0,}', '', self.orisql.strip() + '\n')
        sql = sql.strip()
        sql = sql if sql and sql.endswith(';') else sql + ';' if sql else ''
        return sql

    @property
    def comment(self):
        orisql = self.orisql if self.orisql.endswith(';') else self.orisql + ';'
        comment = orisql.replace(self.sql, '')
        comment = re.search('-- *(.*?)$', comment.strip())
        comment = comment.group(1) if comment else ''
        return comment.strip()

    @property
    def action(self):
        if not self.sql:
            return 'NoSql'

        sql = self.sql.replace('\n', '')
        splits = sql.split(' ')
        action = splits[0]
        if splits[1].lower() == 'index':
            action += splits[1]
        return action.upper()

    @property
    def tablename(self):
        sql = self.sql
        createt = re.search(rf'create table (?:if exists |if not exists )?(.*?){REG_BEHIND}', sql)
        updatet = re.search(rf'update (.*?){REG_BEHIND}', sql)
        deletet = re.search(rf'delete (?:from )?(.*?){REG_BEHIND}', sql)
        insertt = re.search(rf'insert into (.*?){REG_BEHIND}', sql)
        ont = re.search(rf'(?<=on )(.*?){REG_BEHIND}', sql)
        fromt = re.search(rf'select .*? (?<=from )(.*?){REG_BEHIND}', sql, re.S)
        # print(createt, updatet, insertt, ont, fromt)
        tablename = createt or updatet or deletet or insertt or fromt or ont
        tablename = tablename.group(1) if tablename else sql
        return tablename


class SqlFileParse(object):

    def __init__(self, filepath):
        self.filepath = filepath

    def get_content(self):
        if not os.path.isfile(self.filepath):
            raise Exception(f'File 【{self.filepath}】 not exists !')

        with open(self.filepath, 'r', encoding='utf-8') as f:
            content = f.read()
        return content

    def parse_argument(self, argument, arguments):
        key, value = argument.split('=', 1)
        key, value = key.strip(), value.strip()
        try:
            value = eval(value, globals(), arguments)
        except NameError as e:
            raise NameError(f"{e}, please set it before '{key}' !!!")
        return key, value

    @property
    def parameters(self):
        content = self.get_content()
        content = re.sub('--.*?\n', '\n', content)  # 去掉注释
        parameters = re.findall(rf"\$(\w+){REG_BEHIND}", content)
        return set(parameters)

    @property
    def arguments(self):
        '''[summary]

        [description]
            获取文件中配置的arguments
        Returns:
            [dict] -- [返回文件中的参数设置]
        '''
        arguments = {
            'today': date.today(),
            'now': datetime.now(),
        }
        content = self.get_content()
        content = re.sub('--.*?\n', '\n', content)  # 去掉注释
        arguments_infile = re.findall(r'(?<!--)\s*#【arguments】#\s*\n(.*?)#【arguments】#', content, re.S)
        arguments_infile = ';'.join(arguments_infile).replace('\n', ';')
        arguments_infile = [argument.strip() for argument in arguments_infile.split(';') if argument]
        for argument in arguments_infile:
            key, value = self.parse_argument(argument, arguments)
            arguments[key] = value

        arguments = {k: f"'{datetime.strftime(v, '%Y-%m-%d %H:%M:%S')}'" if isinstance(v, datetime)
                        else f"'{datetime.strftime(v, '%Y-%m-%d')}'" if isinstance(v, date)
                        else v for k, v in arguments.items()}  # 处理时间
        return arguments

    def replace_params(self, **kwargs):
        '''[summary]

        [description]
            替换具体的参数值，传入的参数值会覆盖文件中设置的参数值
        Arguments:
            **kwargs {[参数]} -- [传入的参数值]

        Returns:
            [str] -- [替换过后的内容]

        Raises:
            Exception -- [需要设置参数]
        '''
        filename = os.path.basename(self.filepath)
        kwargs = {k: f"'{v}'" if isinstance(v, str) else v for k, v in kwargs.items()}  # str加引号处理
        arguments = self.arguments

        arguments_same = set(arguments) & set(kwargs)
        argsamelog = None
        if arguments_same:
            arguments_same = sorted(arguments_same)
            file_arg = {arg: arguments.get(arg) for arg in arguments_same}
            argsamelog = f"Replace FileSetting {file_arg}"

        arguments.update(kwargs)
        arguments_lack = self.parameters - set(arguments)
        if arguments_lack:
            raise Exception(f"Need params 【{'】, 【'.join(arguments_lack)}】 !")

        content = self.get_content()
        for key, value in arguments.items():
            content = re.sub(rf"\${key}{REG_BEHIND}", f"{value}", content)
        arguments = {k: arguments.get(k) for k in self.parameters}

        arglog = f"【Final Arguments】【{filename}】 Use arguments {arguments}" \
                 if arguments else f"【Final Arguments】【{filename}】 no arguments !!!"
        arglog = arglog + f", {argsamelog}" if argsamelog else arglog
        sqllogger.warning(arglog)

        return arguments, content

    def get_filesqls(self, **kwargs):
        sqls = {}
        arguments, content = self.replace_params(**kwargs)
        sqls_tmp = re.findall(r'(?<!--)\s*###\s*\n(.*?)###', content, re.S)
        for idx, sql in enumerate(sqls_tmp):
            sqlparser = SqlParse(sql)
            purpose = f"【{idx+1:0>3d}】{sqlparser.purpose}"
            sql = re.sub('--【.*?\n', '', sql.strip())
            sqls[purpose] = sql
        return arguments, sqls
